package Ex_1;

class Entry {
	/*@ readonly @*/ Object element;
	/*@ peer @*/ Entry previous, next;

    Entry( /*@ readonly @*/ Object o, /*@ peer @*/ Entry p, /*@ peer @*/ Entry n ) {
	element = o;
	previous = p;
	next = n;
    }
}

public class LinkedList {
    private /*@ rep @*/ Entry header;
    private int size;

    public LinkedList() {
	header = new /*@ rep @*/ Entry(null, null, null);
	header.next = header;
	header.previous = header;

	size = 0;
    }

    public void add( /*@ readonly @*/ Object o ) {
    	/*@ rep @*/ Entry newE = new /*@ rep @*/ Entry( o, header, header.next );

	header.next.previous = newE;
	header.next = newE;

	++size;
    }

    public /*@ pure readonly @*/ Object get( int idx ) {
	if( idx > size ) return null;

	/*@ readonly @*/ Entry e = header.next;

	for( int i=0; i<idx; ++i ) {
	    e = e.next;
	}

	return e.element;
    }

    // precondition that the node belongs to us
    /*@ requires e.owner == this; @*/
    protected void delete( /*@ readonly @*/ Entry e ) {
    /*@ rep @*/ Entry re = (/*@ rep @*/ Entry) e;

	if( re.previous != null )
	    re.previous.next = re.next;

	if( re.next != null )
	    re.next.previous = re.previous;
    }

    public /*@ pure peer @*/ ReadIterator getReadIterator() {
	return new /*@ peer @*/ ReadIterator( header );
    }

    public /*@ pure peer @*/ DeleteIterator getDeleteIterator() {
	return new /*@ peer @*/ DeleteIterator( this, header );
    }


    public static void main( String[] args ) {

	LinkedList ll = new LinkedList();

	ll.add( new Integer(20) );
	ll.add( new String("xyz") );
	ll.add( new Float(2.2f) );
	ll.add( new String("Hello World") );
	ll.add( new Object() );

	ReadIterator itr = ll.getReadIterator();

	int i = 0;
	while( itr.hasNext() ) {
	    itr.moveNext();
	    System.out.println("Element[" + i + "]: " + itr.element() );
	    ++i;
	}

	DeleteIterator itd = ll.getDeleteIterator();
	itd.moveNext();
	itd.moveNext();
	itd.delete();
	System.out.println("Deleted second element.");
	
	itr = ll.getReadIterator();

	i = 0;
	while( itr.hasNext() ) {
	    itr.moveNext();
	    System.out.println("Element[" + i + "]: " + itr.element() );
	    ++i;
	}

    }

}

class ReadIterator {

    public ReadIterator( /*@ readonly @*/ Entry h ) {
	// the header is a dummy that is never null
	current = h;
	header = h;
    }
    
    public /*@ pure @*/ boolean hasNext() {
	return current.next != header;
    }

    public void moveNext() {
	current = current.next;
    }

    public /*@ pure readonly @*/ Object element() {
	return current.element;
    }

    protected /*@ readonly @*/ Entry current;
    protected /*@ readonly @*/ Entry header;

}

class DeleteIterator extends ReadIterator {

    public DeleteIterator( /*@ peer @*/ LinkedList l, /*@ readonly @*/ Entry h ) {
	super( h );
	list = l;
    }
    
    public void delete() {
	list.delete(current);
	current = current.next;
    }

    private /*@ peer @*/ LinkedList list;
}